/*
 * Copyright (c) 2010 Sascha Hauer <s.hauer@pengutronix.de>
 * Copyright (C) 2005-2009 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/types.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/err.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/mfd/imx-ipu-v3.h>
#include <linux/clk.h>
#include <linux/list.h>
#include <linux/module.h>
#include <mach/common.h>
#include <linux/mfd/core.h>
#include <mach/ipu-v3.h>

#include "ipu-prv.h"

static struct clk *ipu_clk;
static struct device *ipu_dev;

static DEFINE_SPINLOCK(ipu_lock);
static DEFINE_MUTEX(ipu_channel_lock);
void __iomem *ipu_cm_reg;
void __iomem *ipu_idmac_reg;

static int ipu_use_count;

static struct ipu_channel channels[64];

struct ipu_channel *ipu_idmac_get(unsigned num)
{
	struct ipu_channel *channel;

	dev_dbg(ipu_dev, "%s %d\n", __func__, num);

	if (num > 63)
		return ERR_PTR(-ENODEV);

	mutex_lock(&ipu_channel_lock);

	channel = &channels[num];

	if (channel->busy) {
		channel = ERR_PTR(-EBUSY);
		goto out;
	}

	channel->busy = 1;
	channel->num = num;

out:
	mutex_unlock(&ipu_channel_lock);

	return channel;
}
EXPORT_SYMBOL(ipu_idmac_get);

void ipu_idmac_put(struct ipu_channel *channel)
{
	dev_dbg(ipu_dev, "%s %d\n", __func__, channel->num);

	mutex_lock(&ipu_channel_lock);

	channel->busy = 0;

	mutex_unlock(&ipu_channel_lock);
}
EXPORT_SYMBOL(ipu_idmac_put);

void ipu_idmac_set_double_buffer(struct ipu_channel *channel, bool doublebuffer)
{
	unsigned long flags;
	u32 reg;

	spin_lock_irqsave(&ipu_lock, flags);

	reg = ipu_cm_read(IPU_CHA_DB_MODE_SEL(channel->num));
	if (doublebuffer)
		reg |= idma_mask(channel->num);
	else
		reg &= ~idma_mask(channel->num);
	ipu_cm_write(reg, IPU_CHA_DB_MODE_SEL(channel->num));

	spin_unlock_irqrestore(&ipu_lock, flags);
}
EXPORT_SYMBOL(ipu_idmac_set_double_buffer);

int ipu_module_enable(u32 mask)
{
	unsigned long lock_flags;
	u32 ipu_conf;

	spin_lock_irqsave(&ipu_lock, lock_flags);

	ipu_conf = ipu_cm_read(IPU_CONF);
	ipu_conf |= mask;
	ipu_cm_write(ipu_conf, IPU_CONF);

	spin_unlock_irqrestore(&ipu_lock, lock_flags);

	return 0;
}

int ipu_module_disable(u32 mask)
{
	unsigned long lock_flags;
	u32 ipu_conf;

	spin_lock_irqsave(&ipu_lock, lock_flags);

	ipu_conf = ipu_cm_read(IPU_CONF);
	ipu_conf &= ~mask;
	ipu_cm_write(ipu_conf, IPU_CONF);

	spin_unlock_irqrestore(&ipu_lock, lock_flags);

	return 0;
}

void ipu_idmac_select_buffer(struct ipu_channel *channel, u32 buf_num)
{
	unsigned int chno = channel->num;
	unsigned long flags;

	spin_lock_irqsave(&ipu_lock, flags);

	/* Mark buffer as ready. */
	if (buf_num == 0)
		ipu_cm_write(idma_mask(chno), IPU_CHA_BUF0_RDY(chno));
	else
		ipu_cm_write(idma_mask(chno), IPU_CHA_BUF1_RDY(chno));

	spin_unlock_irqrestore(&ipu_lock, flags);
}
EXPORT_SYMBOL(ipu_idmac_select_buffer);

int ipu_idmac_enable_channel(struct ipu_channel *channel)
{
	u32 val;
	unsigned long flags;

	ipu_get();

	spin_lock_irqsave(&ipu_lock, flags);

	val = ipu_idmac_read(IDMAC_CHA_EN(channel->num));
	val |= idma_mask(channel->num);
	ipu_idmac_write(val, IDMAC_CHA_EN(channel->num));

	spin_unlock_irqrestore(&ipu_lock, flags);

	return 0;
}
EXPORT_SYMBOL(ipu_idmac_enable_channel);

int ipu_idmac_disable_channel(struct ipu_channel *channel)
{
	u32 val;
	unsigned long flags;

	spin_lock_irqsave(&ipu_lock, flags);

	/* Disable DMA channel(s) */
	val = ipu_idmac_read(IDMAC_CHA_EN(channel->num));
	val &= ~idma_mask(channel->num);
	ipu_idmac_write(val, IDMAC_CHA_EN(channel->num));

	/* Set channel buffers NOT to be ready */
	ipu_cm_write(0xf0000000, IPU_GPR); /* write one to clear */

	if (ipu_idma_is_set(IPU_CHA_BUF0_RDY, channel->num)) {
		ipu_cm_write(idma_mask(channel->num),
			     IPU_CHA_BUF0_RDY(channel->num));
	}
	if (ipu_idma_is_set(IPU_CHA_BUF1_RDY, channel->num)) {
		ipu_cm_write(idma_mask(channel->num),
			     IPU_CHA_BUF1_RDY(channel->num));
	}

	ipu_cm_write(0x0, IPU_GPR); /* write one to set */

	/* Reset the double buffer */
	val = ipu_cm_read(IPU_CHA_DB_MODE_SEL(channel->num));
	val &= ~idma_mask(channel->num);
	ipu_cm_write(val, IPU_CHA_DB_MODE_SEL(channel->num));

	spin_unlock_irqrestore(&ipu_lock, flags);

	ipu_put();

	return 0;
}
EXPORT_SYMBOL(ipu_idmac_disable_channel);

static LIST_HEAD(ipu_irq_handlers);

static void ipu_irq_update_irq_mask(void)
{
	struct ipu_irq_handler *handler;
	int i;

	DECLARE_IPU_IRQ_BITMAP(irqs);

	bitmap_zero(irqs, IPU_IRQ_COUNT);

	list_for_each_entry(handler, &ipu_irq_handlers, list)
		bitmap_or(irqs, irqs, handler->ipu_irqs, IPU_IRQ_COUNT);

	for (i = 0; i < BITS_TO_LONGS(IPU_IRQ_COUNT); i++)
		ipu_cm_write(irqs[i], IPU_INT_CTRL(i + 1));
}

int ipu_irq_add_handler(struct ipu_irq_handler *ipuirq)
{
	unsigned long flags;

	spin_lock_irqsave(&ipu_lock, flags);

	list_add_tail(&ipuirq->list, &ipu_irq_handlers);
	ipu_irq_update_irq_mask();

	spin_unlock_irqrestore(&ipu_lock, flags);
	return 0;
}
EXPORT_SYMBOL(ipu_irq_add_handler);

void ipu_irq_remove_handler(struct ipu_irq_handler *handler)
{
	unsigned long flags;

	spin_lock_irqsave(&ipu_lock, flags);

	list_del(&handler->list);
	ipu_irq_update_irq_mask();

	spin_unlock_irqrestore(&ipu_lock, flags);
}
EXPORT_SYMBOL(ipu_irq_remove_handler);

int ipu_irq_update_handler(struct ipu_irq_handler *handler,
		unsigned long *bitmap)
{
	unsigned long flags;

	spin_lock_irqsave(&ipu_lock, flags);

	bitmap_copy(handler->ipu_irqs, bitmap, IPU_IRQ_COUNT);
	ipu_irq_update_irq_mask();

	spin_unlock_irqrestore(&ipu_lock, flags);

	return 0;
}
EXPORT_SYMBOL(ipu_irq_update_handler);

static void ipu_completion_handler(unsigned long *bitmask, void *context)
{
	struct completion *completion = context;

	complete(completion);
}

int ipu_wait_for_interrupt(int interrupt, int timeout_ms)
{
	struct ipu_irq_handler handler;
	DECLARE_COMPLETION_ONSTACK(completion);
	int ret;

	bitmap_zero(handler.ipu_irqs, IPU_IRQ_COUNT);
	bitmap_set(handler.ipu_irqs, interrupt, 1);

	ipu_cm_write(1 << (interrupt % 32), IPU_INT_STAT(interrupt / 32 + 1));

	handler.handler = ipu_completion_handler;
	handler.context = &completion;
	ipu_irq_add_handler(&handler);

	ret = wait_for_completion_timeout(&completion,
			msecs_to_jiffies(timeout_ms));

	ipu_irq_remove_handler(&handler);

	if (ret > 0)
		ret = 0;
	else
		ret = -ETIMEDOUT;

	return ret;
}
EXPORT_SYMBOL(ipu_wait_for_interrupt);

static irqreturn_t ipu_irq_handler(int irq, void *desc)
{
	DECLARE_IPU_IRQ_BITMAP(status);
	struct ipu_irq_handler *handler;
	int i;

	for (i = 0; i < BITS_TO_LONGS(IPU_IRQ_COUNT); i++) {
		status[i] = ipu_cm_read(IPU_INT_STAT(i + 1));
		ipu_cm_write(status[i], IPU_INT_STAT(i + 1));
	}

	list_for_each_entry(handler, &ipu_irq_handlers, list) {
		DECLARE_IPU_IRQ_BITMAP(tmp);
		if (bitmap_and(tmp, status, handler->ipu_irqs, IPU_IRQ_COUNT))
			handler->handler(tmp, handler->context);
	}

	return IRQ_HANDLED;
}

ipu_color_space_t format_to_colorspace(u32 fmt)
{
	switch (fmt) {
	case IPU_PIX_FMT_RGB666:
	case IPU_PIX_FMT_RGB565:
	case IPU_PIX_FMT_BGR24:
	case IPU_PIX_FMT_RGB24:
	case IPU_PIX_FMT_BGR32:
	case IPU_PIX_FMT_BGRA32:
	case IPU_PIX_FMT_RGB32:
	case IPU_PIX_FMT_RGBA32:
	case IPU_PIX_FMT_ABGR32:
	case IPU_PIX_FMT_LVDS666:
	case IPU_PIX_FMT_LVDS888:
		return RGB;

	default:
		return YCbCr;
	}
}

static int ipu_reset(void)
{
	int timeout = 10000;
	u32 val;

	/* hard reset the IPU */
	val = readl(MX51_IO_ADDRESS(MX51_SRC_BASE_ADDR));
	val |= 1 << 3;
	writel(val, MX51_IO_ADDRESS(MX51_SRC_BASE_ADDR));

	ipu_cm_write(0x807FFFFF, IPU_MEM_RST);

	while (ipu_cm_read(IPU_MEM_RST) & 0x80000000) {
		if (!timeout--)
			return -ETIME;
		udelay(100);
	}

	return 0;
}

/*
 * The MIPI HSC unit has been removed from the i.MX51 Reference Manual by
 * the Freescale marketing division. However this did not remove the
 * hardware from the chip which still needs to be configured...
 */
static int __devinit ipu_mipi_setup(void)
{
	struct clk *hsc_clk;
	void __iomem *hsc_addr;
	int ret = 0;

	hsc_addr = ioremap(MX51_MIPI_HSC_BASE_ADDR, PAGE_SIZE);
	if (!hsc_addr)
		return -ENOMEM;

	hsc_clk = clk_get_sys(NULL, "mipi_hsp");
	if (IS_ERR(hsc_clk)) {
		ret = PTR_ERR(hsc_clk);
		goto unmap;
	}
	clk_enable(hsc_clk);

	/* setup MIPI module to legacy mode */
	__raw_writel(0xF00, hsc_addr);

	/* CSI mode: reserved; DI control mode: legacy (from Freescale BSP) */
	__raw_writel(__raw_readl(hsc_addr + 0x800) | 0x30ff,
		hsc_addr + 0x800);

	clk_disable(hsc_clk);
	clk_put(hsc_clk);
unmap:
	iounmap(hsc_addr);

	return ret;
}

static int ipu_submodules_init(struct platform_device *pdev,
		unsigned long ipu_base, struct clk *ipu_clk)
{
	char *unit;
	int ret;

	ret = ipu_di_init(pdev, 0, ipu_base + IPU_DI0_REG_BASE,
			IPU_CONF_DI0_EN, ipu_clk);
	if (ret) {
		unit = "di0";
		goto err_di_0;
	}

	ret = ipu_di_init(pdev, 1, ipu_base + IPU_DI1_REG_BASE,
			IPU_CONF_DI1_EN, ipu_clk);
	if (ret) {
		unit = "di1";
		goto err_di_1;
	}

	ret = ipu_dc_init(pdev, ipu_base + IPU_DC_REG_BASE,
			ipu_base + IPU_DC_TMPL_REG_BASE);
	if (ret) {
		unit = "dc_template";
		goto err_dc;
	}

	ret = ipu_dmfc_init(pdev, ipu_base + IPU_DMFC_REG_BASE, ipu_clk);
	if (ret) {
		unit = "dmfc";
		goto err_dmfc;
	}

	ret = ipu_dp_init(pdev, ipu_base + IPU_SRM_REG_BASE);
	if (ret) {
		unit = "dp";
		goto err_dp;
	}

	ret = ipu_cpmem_init(pdev, ipu_base + IPU_CPMEM_REG_BASE);
	if (ret) {
		unit = "cpmem";
		goto err_cpmem;
	}

	return 0;

err_cpmem:
	ipu_dp_exit(pdev);
err_dp:
	ipu_dmfc_exit(pdev);
err_dmfc:
	ipu_dc_exit(pdev);
err_dc:
	ipu_di_exit(pdev, 1);
err_di_1:
	ipu_di_exit(pdev, 0);
err_di_0:
	dev_err(&pdev->dev, "init %s failed with %d\n", unit, ret);
	return ret;
}

void ipu_get(void)
{
	mutex_lock(&ipu_channel_lock);

	ipu_use_count++;

	if (ipu_use_count == 1)
		clk_enable(ipu_clk);

	mutex_unlock(&ipu_channel_lock);
}

void ipu_put(void)
{
	mutex_lock(&ipu_channel_lock);

	ipu_use_count--;

	if (ipu_use_count == 0)
		clk_disable(ipu_clk);

	if (ipu_use_count < 0) {
		dev_err(ipu_dev, "ipu use count < 0\n");
		ipu_use_count = 0;
	}

	mutex_unlock(&ipu_channel_lock);
}

static void ipu_submodules_exit(struct platform_device *pdev,
		unsigned long ipu_base)
{
	ipu_cpmem_exit(pdev);
	ipu_dp_exit(pdev);
	ipu_dmfc_exit(pdev);
	ipu_dc_exit(pdev);
	ipu_di_exit(pdev, 1);
	ipu_di_exit(pdev, 0);
}

static int ipu_add_subdevice_pdata(struct platform_device *pdev,
		const char *name, int id, void *pdata)
{
	struct mfd_cell cell = {
		.platform_data = pdata,
		.pdata_size = sizeof(struct ipuv3_fb_platform_data),
	};

	cell.name = name;

	return mfd_add_devices(&pdev->dev, id, &cell, 1, NULL, 0);
}

static int ipu_add_client_devices(struct platform_device *pdev)
{
	struct imx_ipuv3_platform_data *plat_data = pdev->dev.platform_data;
	struct ipuv3_fb_platform_data *fbdata;

	fbdata = plat_data->fb_head0_platform_data;
	if (fbdata) {
		fbdata->ipu_channel_bg =
			MX51_IPU_CHANNEL_MEM_BG_SYNC;
		fbdata->ipu_channel_fg =
			MX51_IPU_CHANNEL_MEM_FG_SYNC;
		fbdata->dc_channel = 5;
		fbdata->dp_channel = IPU_DP_FLOW_SYNC;

		ipu_add_subdevice_pdata(pdev, "imx-ipuv3-fb", 0,
				fbdata);
	}

	fbdata = plat_data->fb_head1_platform_data;
	if (fbdata) {
		fbdata->ipu_channel_bg =
			MX51_IPU_CHANNEL_MEM_DC_SYNC;
		fbdata->ipu_channel_fg = -1;
		fbdata->dc_channel = 1;
		fbdata->dp_channel = -1;

		ipu_add_subdevice_pdata(pdev, "imx-ipuv3-fb", 1,
				fbdata);
	}

	return 0;
}

static int __devinit ipu_probe(struct platform_device *pdev)
{
	struct resource *res;
	unsigned long ipu_base;
	int ret, irq1, irq2;

	/* There can be only one */
	if (ipu_dev)
		return -EBUSY;

	spin_lock_init(&ipu_lock);

	ipu_dev = &pdev->dev;

	irq1 = platform_get_irq(pdev, 0);
	irq2 = platform_get_irq(pdev, 1);
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	if (!res || irq1 < 0 || irq2 < 0)
		return -ENODEV;

	ipu_base = res->start;

	ipu_cm_reg = ioremap(ipu_base + IPU_CM_REG_BASE, PAGE_SIZE);
	if (!ipu_cm_reg) {
		ret = -ENOMEM;
		goto failed_ioremap1;
	}

	ipu_idmac_reg = ioremap(ipu_base + IPU_IDMAC_REG_BASE, PAGE_SIZE);
	if (!ipu_idmac_reg) {
		ret = -ENOMEM;
		goto failed_ioremap2;
	}

	ret = ipu_mipi_setup();
	if (ret)
		goto failed_mipi_setup;

	ipu_clk = clk_get(&pdev->dev, "ipu");
	if (IS_ERR(ipu_clk)) {
		ret = PTR_ERR(ipu_clk);
		dev_err(&pdev->dev, "clk_get failed with %d", ret);
		goto failed_clk_get;
	}

	ipu_get();

	ret = request_irq(irq1, ipu_irq_handler, IRQF_DISABLED, pdev->name,
			&pdev->dev);
	if (ret) {
		dev_err(&pdev->dev, "request irq %d failed with: %d\n", irq1, ret);
		goto failed_request_irq1;
	}

	ret = request_irq(irq2, ipu_irq_handler, IRQF_DISABLED, pdev->name,
			&pdev->dev);
	if (ret) {
		dev_err(&pdev->dev, "request irq %d failed with: %d\n", irq2, ret);
		goto failed_request_irq2;
	}

	ipu_reset();

	ret = ipu_submodules_init(pdev, ipu_base, ipu_clk);
	if (ret)
		goto failed_submodules_init;

	/* Set sync refresh channels as high priority */
	ipu_idmac_write(0x18800000, IDMAC_CHA_PRI(0));

	ret = ipu_add_client_devices(pdev);
        if (ret) {
                dev_err(&pdev->dev, "adding client devices failed with %d\n", ret);
		goto failed_add_clients;
        }

	ipu_put();

	return 0;

failed_add_clients:
	ipu_submodules_exit(pdev, ipu_base);
failed_submodules_init:
	free_irq(irq2, &pdev->dev);
failed_request_irq2:
	free_irq(irq1, &pdev->dev);
	ipu_put();
failed_request_irq1:
	clk_put(ipu_clk);
failed_clk_get:
failed_mipi_setup:
	iounmap(ipu_idmac_reg);
failed_ioremap2:
	iounmap(ipu_cm_reg);
failed_ioremap1:

	return ret;
}

static int __devexit ipu_remove(struct platform_device *pdev)
{
	struct resource *res;
	unsigned long ipu_base;
	int irq1, irq2;

	irq1 = platform_get_irq(pdev, 0);
	irq2 = platform_get_irq(pdev, 1);
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	ipu_base = res->start;

	mfd_remove_devices(&pdev->dev);
	ipu_submodules_exit(pdev, ipu_base);
	free_irq(irq2, &pdev->dev);
	free_irq(irq1, &pdev->dev);
	iounmap(ipu_idmac_reg);
	iounmap(ipu_cm_reg);

	if (ipu_use_count != 0) {
		dev_err(ipu_dev, "unbalanced use count: %d\n", ipu_use_count);
		clk_disable(ipu_clk);
	}

	clk_put(ipu_clk);
	ipu_dev = NULL;

	return 0;
}

static struct platform_driver mxcipu_driver = {
	.driver = {
		.name = "imx-ipuv3",
	},
	.probe = ipu_probe,
	.remove = __devexit_p(ipu_remove),
};

static int __init ipu_gen_init(void)
{
	int32_t ret;

	ret = platform_driver_register(&mxcipu_driver);
	return 0;
}
subsys_initcall(ipu_gen_init);

static void __exit ipu_gen_uninit(void)
{
	platform_driver_unregister(&mxcipu_driver);
}
module_exit(ipu_gen_uninit);

MODULE_DESCRIPTION("i.MX IPU v3 driver");
MODULE_AUTHOR("Sascha Hauer <s.hauer@pengutronix.de>");
MODULE_LICENSE("GPL");
