/*
 * Copyright (c) 2010 Sascha Hauer <s.hauer@pengutronix.de>
 * Copyright (C) 2005-2009 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/mfd/imx-ipu-v3.h>
#include <linux/types.h>
#include <linux/bitrev.h>
#include <linux/io.h>
#include <linux/export.h>

#include "ipu-prv.h"

#define __F(word, ofs, size) ((((word) * 160 + (ofs)) << 8) | (size))

#define IPU_FIELD_XV		__F(0, 0, 10)
#define IPU_FIELD_YV		__F(0, 10, 9)
#define IPU_FIELD_XB		__F(0, 19, 13)
#define IPU_FIELD_YB		__F(0, 32, 12)
#define IPU_FIELD_NSB_B		__F(0, 44, 1)
#define IPU_FIELD_CF		__F(0, 45, 1)
#define IPU_FIELD_UBO		__F(0, 46, 22)
#define IPU_FIELD_VBO		__F(0, 68, 22)
#define IPU_FIELD_IOX		__F(0, 90, 4)
#define IPU_FIELD_RDRW		__F(0, 94, 1)
#define IPU_FIELD_SO		__F(0, 113, 1)
#define IPU_FIELD_BNDM		__F(0, 114, 3)
#define IPU_FIELD_BM		__F(0, 117, 2)
#define IPU_FIELD_ROT		__F(0, 119, 1)
#define IPU_FIELD_HF		__F(0, 120, 1)
#define IPU_FIELD_VF		__F(0, 121, 1)
#define IPU_FIELD_THE		__F(0, 122, 1)
#define IPU_FIELD_CAP		__F(0, 123, 1)
#define IPU_FIELD_CAE		__F(0, 124, 1)
#define IPU_FIELD_FW		__F(0, 125, 13)
#define IPU_FIELD_FH		__F(0, 138, 12)
#define IPU_FIELD_EBA0		__F(1, 0, 29)
#define IPU_FIELD_EBA1		__F(1, 29, 29)
#define IPU_FIELD_ILO		__F(1, 58, 20)
#define IPU_FIELD_NPB		__F(1, 78, 7)
#define IPU_FIELD_PFS		__F(1, 85, 4)
#define IPU_FIELD_ALU		__F(1, 89, 1)
#define IPU_FIELD_ALBM		__F(1, 90, 3)
#define IPU_FIELD_ID		__F(1, 93, 2)
#define IPU_FIELD_TH		__F(1, 95, 7)
#define IPU_FIELD_SLY		__F(1, 102, 14)
#define IPU_FIELD_WID3		__F(1, 125, 3)
#define IPU_FIELD_SLUV		__F(1, 128, 14)
#define IPU_FIELD_CRE		__F(1, 149, 1)

#define IPU_FIELD_XV		__F(0, 0, 10)
#define IPU_FIELD_YV		__F(0, 10, 9)
#define IPU_FIELD_XB		__F(0, 19, 13)
#define IPU_FIELD_YB		__F(0, 32, 12)
#define IPU_FIELD_NSB_B		__F(0, 44, 1)
#define IPU_FIELD_CF		__F(0, 45, 1)
#define IPU_FIELD_SX		__F(0, 46, 12)
#define IPU_FIELD_SY		__F(0, 58, 11)
#define IPU_FIELD_NS		__F(0, 69, 10)
#define IPU_FIELD_SDX		__F(0, 79, 7)
#define IPU_FIELD_SM		__F(0, 86, 10)
#define IPU_FIELD_SCC		__F(0, 96, 1)
#define IPU_FIELD_SCE		__F(0, 97, 1)
#define IPU_FIELD_SDY		__F(0, 98, 7)
#define IPU_FIELD_SDRX		__F(0, 105, 1)
#define IPU_FIELD_SDRY		__F(0, 106, 1)
#define IPU_FIELD_BPP		__F(0, 107, 3)
#define IPU_FIELD_DEC_SEL	__F(0, 110, 2)
#define IPU_FIELD_DIM		__F(0, 112, 1)
#define IPU_FIELD_SO		__F(0, 113, 1)
#define IPU_FIELD_BNDM		__F(0, 114, 3)
#define IPU_FIELD_BM		__F(0, 117, 2)
#define IPU_FIELD_ROT		__F(0, 119, 1)
#define IPU_FIELD_HF		__F(0, 120, 1)
#define IPU_FIELD_VF		__F(0, 121, 1)
#define IPU_FIELD_THE		__F(0, 122, 1)
#define IPU_FIELD_CAP		__F(0, 123, 1)
#define IPU_FIELD_CAE		__F(0, 124, 1)
#define IPU_FIELD_FW		__F(0, 125, 13)
#define IPU_FIELD_FH		__F(0, 138, 12)
#define IPU_FIELD_EBA0		__F(1, 0, 29)
#define IPU_FIELD_EBA1		__F(1, 29, 29)
#define IPU_FIELD_ILO		__F(1, 58, 20)
#define IPU_FIELD_NPB		__F(1, 78, 7)
#define IPU_FIELD_PFS		__F(1, 85, 4)
#define IPU_FIELD_ALU		__F(1, 89, 1)
#define IPU_FIELD_ALBM		__F(1, 90, 3)
#define IPU_FIELD_ID		__F(1, 93, 2)
#define IPU_FIELD_TH		__F(1, 95, 7)
#define IPU_FIELD_SL		__F(1, 102, 14)
#define IPU_FIELD_WID0		__F(1, 116, 3)
#define IPU_FIELD_WID1		__F(1, 119, 3)
#define IPU_FIELD_WID2		__F(1, 122, 3)
#define IPU_FIELD_WID3		__F(1, 125, 3)
#define IPU_FIELD_OFS0		__F(1, 128, 5)
#define IPU_FIELD_OFS1		__F(1, 133, 5)
#define IPU_FIELD_OFS2		__F(1, 138, 5)
#define IPU_FIELD_OFS3		__F(1, 143, 5)
#define IPU_FIELD_SXYS		__F(1, 148, 1)
#define IPU_FIELD_CRE		__F(1, 149, 1)
#define IPU_FIELD_DEC_SEL2	__F(1, 150, 1)

static u32 *ipu_cpmem_base;
static struct device *ipu_dev;

struct ipu_ch_param_word {
	u32 data[5];
	u32 res[3];
};

struct ipu_ch_param {
	struct ipu_ch_param_word word[2];
};

static u32 ipu_bytes_per_pixel(u32 fmt)
{
	switch (fmt) {
	case IPU_PIX_FMT_GENERIC:	/* generic data */
	case IPU_PIX_FMT_RGB332:
	case IPU_PIX_FMT_YUV420P:
	case IPU_PIX_FMT_YUV422P:
		return 1;

	case IPU_PIX_FMT_RGB565:
	case IPU_PIX_FMT_YUYV:
	case IPU_PIX_FMT_UYVY:
		return 2;

	case IPU_PIX_FMT_BGR24:
	case IPU_PIX_FMT_RGB24:
		return 3;

	case IPU_PIX_FMT_GENERIC_32:	/* generic data */
	case IPU_PIX_FMT_BGR32:
	case IPU_PIX_FMT_BGRA32:
	case IPU_PIX_FMT_RGB32:
	case IPU_PIX_FMT_RGBA32:
	case IPU_PIX_FMT_ABGR32:
		return 4;

	default:
		return 1;
	}
}

bool ipu_pixel_format_has_alpha(u32 fmt)
{
	switch (fmt) {
	case IPU_PIX_FMT_RGBA32:
	case IPU_PIX_FMT_BGRA32:
	case IPU_PIX_FMT_ABGR32:
		return true;

	default:
		return false;
	}
}

#define ipu_ch_param_addr(ch) (((struct ipu_ch_param *)ipu_cpmem_base) + (ch))

static inline void ipu_ch_param_set_field(struct ipu_ch_param *base, u32 wbs, u32 v)
{
	u32 bit = (wbs >> 8) % 160;
	u32 size = wbs & 0xff;
	u32 word = (wbs >> 8) / 160;
	u32 i = bit / 32;
	u32 ofs = bit % 32;
	u32 mask = (1 << size) - 1;

	pr_debug("%s %d %d %d\n", __func__, word, bit , size);

	base->word[word].data[i] &= ~(mask << ofs);
	base->word[word].data[i] |= v << ofs;

	if ((bit + size - 1) / 32 > i) {
		base->word[word].data[i + 1] &= ~(v >> (mask ? (32 - ofs) : 0));
		base->word[word].data[i + 1] |= v >> (ofs ? (32 - ofs) : 0);
	}
}

static inline u32 ipu_ch_param_read_field(struct ipu_ch_param *base, u32 wbs)
{
	u32 bit = (wbs >> 8) % 160;
	u32 size = wbs & 0xff;
	u32 word = (wbs >> 8) / 160;
	u32 i = bit / 32;
	u32 ofs = bit % 32;
	u32 mask = (1 << size) - 1;
	u32 val = 0;

	pr_debug("%s %d %d %d\n", __func__, word, bit , size);

	val = (base->word[word].data[i] >> ofs) & mask;

	if ((bit + size - 1) / 32 > i) {
		u32 tmp;
		tmp = base->word[word].data[i + 1];
		tmp &= mask >> (ofs ? (32 - ofs) : 0);
		val |= tmp << (ofs ? (32 - ofs) : 0);
	}

	return val;
}

static inline void ipu_ch_params_set_packing(struct ipu_ch_param *p,
					      int red_width, int red_offset,
					      int green_width, int green_offset,
					      int blue_width, int blue_offset,
					      int alpha_width, int alpha_offset)
{
	/* Setup red width and offset */
	ipu_ch_param_set_field(p, IPU_FIELD_WID0, red_width - 1);
	ipu_ch_param_set_field(p, IPU_FIELD_OFS0, red_offset);
	/* Setup green width and offset */
	ipu_ch_param_set_field(p, IPU_FIELD_WID1, green_width - 1);
	ipu_ch_param_set_field(p, IPU_FIELD_OFS1, green_offset);
	/* Setup blue width and offset */
	ipu_ch_param_set_field(p, IPU_FIELD_WID2, blue_width - 1);
	ipu_ch_param_set_field(p, IPU_FIELD_OFS2, blue_offset);
	/* Setup alpha width and offset */
	ipu_ch_param_set_field(p, IPU_FIELD_WID3, alpha_width - 1);
	ipu_ch_param_set_field(p, IPU_FIELD_OFS3, alpha_offset);
}

static inline void ipu_ch_param_dump(int ch)
{
	struct ipu_ch_param *p = ipu_ch_param_addr(ch);
	pr_debug("ch %d word 0 - %08X %08X %08X %08X %08X\n", ch,
		 p->word[0].data[0], p->word[0].data[1], p->word[0].data[2],
		 p->word[0].data[3], p->word[0].data[4]);
	pr_debug("ch %d word 1 - %08X %08X %08X %08X %08X\n", ch,
		 p->word[1].data[0], p->word[1].data[1], p->word[1].data[2],
		 p->word[1].data[3], p->word[1].data[4]);
	pr_debug("PFS 0x%x\n", ipu_ch_param_read_field(p, IPU_FIELD_PFS));
	pr_debug("BPP 0x%x\n", ipu_ch_param_read_field(p, IPU_FIELD_BPP));
	pr_debug("NPB 0x%x\n", ipu_ch_param_read_field(p, IPU_FIELD_NPB));

	pr_debug("FW %d\n", ipu_ch_param_read_field(p, IPU_FIELD_FW));
	pr_debug("FH %d\n", ipu_ch_param_read_field(p, IPU_FIELD_FH));
	pr_debug("Stride %d\n", ipu_ch_param_read_field(p, IPU_FIELD_SL));

	pr_debug("Width0 %d+1\n", ipu_ch_param_read_field(p, IPU_FIELD_WID0));
	pr_debug("Width1 %d+1\n", ipu_ch_param_read_field(p, IPU_FIELD_WID1));
	pr_debug("Width2 %d+1\n", ipu_ch_param_read_field(p, IPU_FIELD_WID2));
	pr_debug("Width3 %d+1\n", ipu_ch_param_read_field(p, IPU_FIELD_WID3));
	pr_debug("Offset0 %d\n", ipu_ch_param_read_field(p, IPU_FIELD_OFS0));
	pr_debug("Offset1 %d\n", ipu_ch_param_read_field(p, IPU_FIELD_OFS1));
	pr_debug("Offset2 %d\n", ipu_ch_param_read_field(p, IPU_FIELD_OFS2));
	pr_debug("Offset3 %d\n", ipu_ch_param_read_field(p, IPU_FIELD_OFS3));
}

static inline void ipu_ch_param_set_burst_size(u32 ch,
						u16 burst_pixels)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_NPB,
			       burst_pixels - 1);
};

static inline int ipu_ch_param_get_burst_size(u32 ch)
{
	return ipu_ch_param_read_field(ipu_ch_param_addr(ch), IPU_FIELD_NPB) + 1;
};

static inline int ipu_ch_param_get_bpp(u32 ch)
{
	return ipu_ch_param_read_field(ipu_ch_param_addr(ch), IPU_FIELD_BPP);
};

static inline void ipu_ch_param_set_buffer(u32 ch, int bufNum,
					    dma_addr_t phyaddr)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch),
			bufNum ? IPU_FIELD_EBA1 : IPU_FIELD_EBA0,
			phyaddr / 8);
};

#define IPU_FIELD_ROT_HF_VF		__F(0, 119, 3)

static inline void ipu_ch_param_set_rotation(u32 ch,
					      ipu_rotate_mode_t rot)
{
	u32 temp_rot = bitrev8(rot) >> 5;
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ROT_HF_VF, temp_rot);
};

static inline void ipu_ch_param_set_block_mode(u32 ch)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_BM, 1);
};

static inline void ipu_ch_param_set_alpha_use_separate_channel(u32 ch,
		bool option)
{
	if (option)
		ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ALU, 1);
	else
		ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ALU, 0);
};

static inline void ipu_ch_param_set_alpha_condition_read(u32 ch)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_CRE, 1);
};

static inline void ipu_ch_param_set_alpha_buffer_memory(u32 ch)
{
	int alp_mem_idx;

	switch (ch) {
	case 14: /* PRP graphic */
		alp_mem_idx = 0;
		break;
	case 15: /* PP graphic */
		alp_mem_idx = 1;
		break;
	case 23: /* DP BG SYNC graphic */
		alp_mem_idx = 4;
		break;
	case 27: /* DP FG SYNC graphic */
		alp_mem_idx = 2;
		break;
	default:
		dev_err(ipu_dev, "unsupported correlated channel of local alpha channel\n");
		return;
	}

	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ALBM, alp_mem_idx);
};

static inline void ipu_ch_param_set_interlaced_scan(u32 ch)
{
	u32 stride;
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_SO, 1);
	stride = ipu_ch_param_read_field(ipu_ch_param_addr(ch), IPU_FIELD_SL) + 1;
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ILO, stride / 8);
	stride *= 2;
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_SLY, stride - 1);
};

static inline void ipu_ch_param_set_high_priority(u32 ch)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_ID, 1);
};

static inline void ipu_ch_params_set_alpha_width(u32 ch, int alpha_width)
{
	ipu_ch_param_set_field(ipu_ch_param_addr(ch), IPU_FIELD_WID3,
			alpha_width - 1);
}

static int ipu_ch_param_init(int ch,
			u32 pixel_fmt, u32 width,
			u32 height, u32 stride,
			u32 u, u32 v,
			u32 uv_stride, dma_addr_t addr0,
			dma_addr_t addr1)
{
	u32 u_offset = 0;
	u32 v_offset = 0;
	struct ipu_ch_param params;

	memset(&params, 0, sizeof(params));

	ipu_ch_param_set_field(&params, IPU_FIELD_FW, width - 1);
	ipu_ch_param_set_field(&params, IPU_FIELD_FH, height - 1);
	ipu_ch_param_set_field(&params, IPU_FIELD_SLY, stride - 1);
	ipu_ch_param_set_field(&params, IPU_FIELD_EBA0, addr0 >> 3);
	ipu_ch_param_set_field(&params, IPU_FIELD_EBA1, addr1 >> 3);

	switch (pixel_fmt) {
	case IPU_PIX_FMT_GENERIC:
		/* Represents 8-bit Generic data */
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 5);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 6);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 63);	/* burst size */

		break;
	case IPU_PIX_FMT_GENERIC_32:
		/* Represents 32-bit Generic data */
		break;
	case IPU_PIX_FMT_RGB565:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 3);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 15);	/* burst size */

		ipu_ch_params_set_packing(&params, 5, 0, 6, 5, 5, 11, 8, 16);
		break;
	case IPU_PIX_FMT_BGR24:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 1);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 19);	/* burst size */

		ipu_ch_params_set_packing(&params, 8, 0, 8, 8, 8, 16, 8, 24);
		break;
	case IPU_PIX_FMT_RGB24:
	case IPU_PIX_FMT_YUV444:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 1);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 19);	/* burst size */

		ipu_ch_params_set_packing(&params, 8, 16, 8, 8, 8, 0, 8, 24);
		break;
	case IPU_PIX_FMT_BGRA32:
	case IPU_PIX_FMT_BGR32:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 0);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 15);	/* burst size */

		ipu_ch_params_set_packing(&params, 8, 8, 8, 16, 8, 24, 8, 0);
		break;
	case IPU_PIX_FMT_RGBA32:
	case IPU_PIX_FMT_RGB32:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 0);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 15);	/* burst size */

		ipu_ch_params_set_packing(&params, 8, 24, 8, 16, 8, 8, 8, 0);
		break;
	case IPU_PIX_FMT_ABGR32:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 0);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 7);	/* pix format */

		ipu_ch_params_set_packing(&params, 8, 0, 8, 8, 8, 16, 8, 24);
		break;
	case IPU_PIX_FMT_UYVY:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 3);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 0xA);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 15);	/* burst size */
		break;
	case IPU_PIX_FMT_YUYV:
		ipu_ch_param_set_field(&params, IPU_FIELD_BPP, 3);	/* bits/pixel */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 0x8);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 31);	/* burst size */
		break;
	case IPU_PIX_FMT_YUV420P2:
	case IPU_PIX_FMT_YUV420P:
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 2);	/* pix format */

		if (uv_stride < stride / 2)
			uv_stride = stride / 2;

		u_offset = stride * height;
		v_offset = u_offset + (uv_stride * height / 2);
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 31);  /* burst size */
		break;
	case IPU_PIX_FMT_YVU422P:
		/* BPP & pixel format */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 1);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 31);	/* burst size */

		if (uv_stride < stride / 2)
			uv_stride = stride / 2;

		v_offset = (v == 0) ? stride * height : v;
		u_offset = (u == 0) ? v_offset + v_offset / 2 : u;
		break;
	case IPU_PIX_FMT_YUV422P:
		/* BPP & pixel format */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 1);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 31);	/* burst size */

		if (uv_stride < stride / 2)
			uv_stride = stride / 2;

		u_offset = (u == 0) ? stride * height : u;
		v_offset = (v == 0) ? u_offset + u_offset / 2 : v;
		break;
	case IPU_PIX_FMT_NV12:
		/* BPP & pixel format */
		ipu_ch_param_set_field(&params, IPU_FIELD_PFS, 4);	/* pix format */
		ipu_ch_param_set_field(&params, IPU_FIELD_NPB, 31);	/* burst size */
		uv_stride = stride;
		u_offset = (u == 0) ? stride * height : u;
		break;
	default:
		dev_err(ipu_dev, "mxc ipu: unimplemented pixel format: %d\n",
			pixel_fmt);
		return -EINVAL;
	}
	/* set burst size to 16 */
	if (uv_stride)
		ipu_ch_param_set_field(&params, IPU_FIELD_SLUV, uv_stride - 1);

	if (u > u_offset)
		u_offset = u;

	if (v > v_offset)
		v_offset = v;

	ipu_ch_param_set_field(&params, IPU_FIELD_UBO, u_offset / 8);
	ipu_ch_param_set_field(&params, IPU_FIELD_VBO, v_offset / 8);

	pr_debug("initializing idma ch %d @ %p\n", ch, ipu_ch_param_addr(ch));
	memcpy(ipu_ch_param_addr(ch), &params, sizeof(params));
	return 0;
}

/*
 * This function is called to initialize a buffer for a IPU channel.
 *
 * @param       channel         The IPU channel.
 *
 * @param       pixel_fmt       Input parameter for pixel format of buffer.
 *                              Pixel format is a FOURCC ASCII code.
 *
 * @param       width           Input parameter for width of buffer in pixels.
 *
 * @param       height          Input parameter for height of buffer in pixels.
 *
 * @param       stride          Input parameter for stride length of buffer
 *                              in pixels.
 *
 * @param       rot_mode        Input parameter for rotation setting of buffer.
 *                              A rotation setting other than
 *                              IPU_ROTATE_VERT_FLIP
 *                              should only be used for input buffers of
 *                              rotation channels.
 *
 * @param       phyaddr_0       Input parameter buffer 0 physical address.
 *
 * @param       phyaddr_1       Input parameter buffer 1 physical address.
 *                              Setting this to a value other than NULL enables
 *                              double buffering mode.
 *
 * @param       u		private u offset for additional cropping,
 *				zero if not used.
 *
 * @param       v		private v offset for additional cropping,
 *				zero if not used.
 *
 * @return      Returns 0 on success or negative error code on fail
 */
int ipu_idmac_init_channel_buffer(struct ipu_channel *channel,
				u32 pixel_fmt,
				u16 width, u16 height,
				u32 stride,
				ipu_rotate_mode_t rot_mode,
				dma_addr_t phyaddr_0, dma_addr_t phyaddr_1,
				u32 u, u32 v, bool interlaced)
{
	int ret = 0;
	u32 dma_chan = channel->num;

	if (stride < width * ipu_bytes_per_pixel(pixel_fmt))
		stride = width * ipu_bytes_per_pixel(pixel_fmt);

	if (stride % 4) {
		dev_err(ipu_dev,
			"Stride not 32-bit aligned, stride = %d\n", stride);
		return -EINVAL;
	}

	ipu_get();

	/* Build parameter memory data for DMA channel */
	ret = ipu_ch_param_init(dma_chan, pixel_fmt, width, height, stride, u, v, 0,
			   phyaddr_0, phyaddr_1);
	if (ret)
		goto out;

	if (rot_mode)
		ipu_ch_param_set_rotation(dma_chan, rot_mode);

	if (interlaced)
		ipu_ch_param_set_interlaced_scan(dma_chan);

	if (idmac_idma_is_set(IDMAC_CHA_PRI, dma_chan))
		ipu_ch_param_set_high_priority(dma_chan);

	ipu_ch_param_dump(dma_chan);
out:
	ipu_put();

	return ret;
}
EXPORT_SYMBOL(ipu_idmac_init_channel_buffer);

int ipu_idmac_update_channel_buffer(struct ipu_channel *channel,
				  u32 buf_num, dma_addr_t phyaddr)
{
	u32 dma_chan = channel->num;

	ipu_ch_param_set_buffer(dma_chan, buf_num, phyaddr);

	return 0;
}
EXPORT_SYMBOL(ipu_idmac_update_channel_buffer);

int ipu_cpmem_init(struct platform_device *pdev, unsigned long base)
{
	ipu_cpmem_base = ioremap(base, PAGE_SIZE);
	if (!ipu_cpmem_base)
		return -ENOMEM;
	ipu_dev = &pdev->dev;
	return 0;
}

void ipu_cpmem_exit(struct platform_device *pdev)
{
	iounmap(ipu_cpmem_base);
}
