/*
 * Copyright (c) 2010 Sascha Hauer <s.hauer@pengutronix.de>
 * Copyright (C) 2005-2009 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/io.h>
#include <linux/export.h>
#include <linux/mfd/imx-ipu-v3.h>

#include "ipu-prv.h"

#define ASYNC_SER_WAVE 6

#define DC_DISP_ID_SERIAL	2
#define DC_DISP_ID_ASYNC	3

#define DC_MAP_CONF_PTR(n)	(ipu_dc_reg + 0x0108 + ((n) & ~0x1) * 2)
#define DC_MAP_CONF_VAL(n)	(ipu_dc_reg + 0x0144 + ((n) & ~0x1) * 2)

#define DC_EVT_NF		0
#define DC_EVT_NL		1
#define DC_EVT_EOF		2
#define DC_EVT_NFIELD		3
#define DC_EVT_EOL		4
#define DC_EVT_EOFIELD		5
#define DC_EVT_NEW_ADDR		6
#define DC_EVT_NEW_CHAN		7
#define DC_EVT_NEW_DATA		8

#define DC_EVT_NEW_ADDR_W_0	0
#define DC_EVT_NEW_ADDR_W_1	1
#define DC_EVT_NEW_CHAN_W_0	2
#define DC_EVT_NEW_CHAN_W_1	3
#define DC_EVT_NEW_DATA_W_0	4
#define DC_EVT_NEW_DATA_W_1	5
#define DC_EVT_NEW_ADDR_R_0	6
#define DC_EVT_NEW_ADDR_R_1	7
#define DC_EVT_NEW_CHAN_R_0	8
#define DC_EVT_NEW_CHAN_R_1	9
#define DC_EVT_NEW_DATA_R_0	10
#define DC_EVT_NEW_DATA_R_1	11

#define DC_WR_CH_CONF(ch)	(ipu_dc_reg + dc_channels[ch].channel_offset)
#define DC_WR_CH_ADDR(ch)	(ipu_dc_reg + dc_channels[ch].channel_offset + 4)
#define DC_RL_CH(ch, evt)	(ipu_dc_reg + dc_channels[ch].channel_offset + 8 + ((evt) & ~0x1) * 2)

#define DC_GEN			(ipu_dc_reg + 0x00D4)
#define DC_DISP_CONF1(disp)	(ipu_dc_reg + 0x00D8 + disp * 4)
#define DC_DISP_CONF2(disp)	(ipu_dc_reg + 0x00E8 + disp * 4)
#define DC_STAT			(ipu_dc_reg + 0x01C8)

#define WROD(lf)		(0x18 | (lf << 1))

#define DC_WR_CH_CONF_FIELD_MODE		(1 << 9)
#define DC_WR_CH_CONF_PROG_TYPE_OFFSET		5
#define DC_WR_CH_CONF_PROG_TYPE_MASK		(7 << 5)
#define DC_WR_CH_CONF_PROG_DI_ID		(1 << 2)
#define DC_WR_CH_CONF_PROG_DISP_ID_OFFSET	3
#define DC_WR_CH_CONF_PROG_DISP_ID_MASK		(3 << 3)

static void __iomem *ipu_dc_reg;
static void __iomem *ipu_dc_tmpl_reg;
static struct device *ipu_dev;

struct ipu_dc {
	unsigned int di; /* The display interface number assigned to this dc channel */
	unsigned int channel_offset;
};

static struct ipu_dc dc_channels[10];

static void ipu_dc_link_event(int chan, int event, int addr, int priority)
{
	u32 reg;

	reg = __raw_readl(DC_RL_CH(chan, event));
	reg &= ~(0xFFFF << (16 * (event & 0x1)));
	reg |= ((addr << 8) | priority) << (16 * (event & 0x1));
	__raw_writel(reg, DC_RL_CH(chan, event));
}

static void ipu_dc_write_tmpl(int word, u32 opcode, u32 operand, int map,
			       int wave, int glue, int sync)
{
	u32 reg;
	int stop = 1;

	reg = sync;
	reg |= (glue << 4);
	reg |= (++wave << 11);
	reg |= (++map << 15);
	reg |= (operand << 20) & 0xFFF00000;
	__raw_writel(reg, ipu_dc_tmpl_reg + word * 8);

	reg = (operand >> 12);
	reg |= opcode << 4;
	reg |= (stop << 9);
	__raw_writel(reg, ipu_dc_tmpl_reg + word * 8 + 4);
}

static int ipu_pixfmt_to_map(u32 fmt)
{
	switch (fmt) {
	case IPU_PIX_FMT_GENERIC:
	case IPU_PIX_FMT_RGB24:
		return 0;
	case IPU_PIX_FMT_RGB666:
		return 1;
	case IPU_PIX_FMT_YUV444:
		return 2;
	case IPU_PIX_FMT_RGB565:
		return 3;
	case IPU_PIX_FMT_LVDS666:
		return 4;
	}

	return -EINVAL;
}

#define SYNC_WAVE 0

int ipu_dc_init_sync(int dc_chan, int di, bool interlaced, u32 pixel_fmt, u32 width)
{
	u32 reg = 0, map;

	dc_channels[dc_chan].di = di;

	map = ipu_pixfmt_to_map(pixel_fmt);
	if (map < 0) {
		dev_dbg(ipu_dev, "IPU_DISP: No MAP\n");
		return -EINVAL;
	}

	ipu_get();

	if (interlaced) {
		ipu_dc_link_event(dc_chan, DC_EVT_NL, 0, 3);
		ipu_dc_link_event(dc_chan, DC_EVT_EOL, 0, 2);
		ipu_dc_link_event(dc_chan, DC_EVT_NEW_DATA, 0, 1);

		/* Init template microcode */
		ipu_dc_write_tmpl(0, WROD(0), 0, map, SYNC_WAVE, 0, 8);
	} else {
		if (di) {
			ipu_dc_link_event(dc_chan, DC_EVT_NL, 2, 3);
			ipu_dc_link_event(dc_chan, DC_EVT_EOL, 3, 2);
			ipu_dc_link_event(dc_chan, DC_EVT_NEW_DATA, 4, 1);
			/* Init template microcode */
			ipu_dc_write_tmpl(2, WROD(0), 0, map, SYNC_WAVE, 8, 5);
			ipu_dc_write_tmpl(3, WROD(0), 0, map, SYNC_WAVE, 4, 5);
			ipu_dc_write_tmpl(4, WROD(0), 0, map, SYNC_WAVE, 0, 5);
		} else {
			ipu_dc_link_event(dc_chan, DC_EVT_NL, 5, 3);
			ipu_dc_link_event(dc_chan, DC_EVT_EOL, 6, 2);
			ipu_dc_link_event(dc_chan, DC_EVT_NEW_DATA, 7, 1);
			/* Init template microcode */
			ipu_dc_write_tmpl(5, WROD(0), 0, map, SYNC_WAVE, 8, 5);
			ipu_dc_write_tmpl(6, WROD(0), 0, map, SYNC_WAVE, 4, 5);
			ipu_dc_write_tmpl(7, WROD(0), 0, map, SYNC_WAVE, 0, 5);
		}
	}
	ipu_dc_link_event(dc_chan, DC_EVT_NF, 0, 0);
	ipu_dc_link_event(dc_chan, DC_EVT_NFIELD, 0, 0);
	ipu_dc_link_event(dc_chan, DC_EVT_EOF, 0, 0);
	ipu_dc_link_event(dc_chan, DC_EVT_EOFIELD, 0, 0);
	ipu_dc_link_event(dc_chan, DC_EVT_NEW_CHAN, 0, 0);
	ipu_dc_link_event(dc_chan, DC_EVT_NEW_ADDR, 0, 0);

	reg = 0x2;
	reg |= di << DC_WR_CH_CONF_PROG_DISP_ID_OFFSET;
	reg |= di << 2;
	if (interlaced)
		reg |= DC_WR_CH_CONF_FIELD_MODE;

	__raw_writel(reg, DC_WR_CH_CONF(dc_chan));

	__raw_writel(0x00000000, DC_WR_CH_ADDR(dc_chan));

	__raw_writel(0x00000084, DC_GEN);

	__raw_writel(width, DC_DISP_CONF2(di));

	ipu_module_enable(IPU_CONF_DC_EN);

	ipu_put();

	return 0;
}
EXPORT_SYMBOL(ipu_dc_init_sync);

void ipu_dc_init_async(int dc_chan, int di, bool interlaced)
{
	u32 reg = 0;
	dc_channels[dc_chan].di = di;
	ipu_dc_link_event(dc_chan, DC_EVT_NEW_DATA_W_0, 0x64, 1);
	ipu_dc_link_event(dc_chan, DC_EVT_NEW_DATA_W_1, 0x64, 1);

	reg = 0x3;
	reg |= DC_DISP_ID_SERIAL << DC_WR_CH_CONF_PROG_DISP_ID_OFFSET;
	__raw_writel(reg, DC_WR_CH_CONF(dc_chan));

	__raw_writel(0x00000000, DC_WR_CH_ADDR(dc_chan));

	__raw_writel(0x00000084, DC_GEN);

	ipu_module_enable(IPU_CONF_DC_EN);
}
EXPORT_SYMBOL(ipu_dc_init_async);

void ipu_dc_enable_channel(u32 dc_chan)
{
	int di;
	u32 reg;

	di = dc_channels[dc_chan].di;

	/* Make sure other DC sync channel is not assigned same DI */
	reg = __raw_readl(DC_WR_CH_CONF(6 - dc_chan));
	if ((di << 2) == (reg & DC_WR_CH_CONF_PROG_DI_ID)) {
		reg &= ~DC_WR_CH_CONF_PROG_DI_ID;
		reg |= di ? 0 : DC_WR_CH_CONF_PROG_DI_ID;
		__raw_writel(reg, DC_WR_CH_CONF(6 - dc_chan));
	}

	reg = __raw_readl(DC_WR_CH_CONF(dc_chan));
	reg |= 4 << DC_WR_CH_CONF_PROG_TYPE_OFFSET;
	__raw_writel(reg, DC_WR_CH_CONF(dc_chan));
}
EXPORT_SYMBOL(ipu_dc_enable_channel);

void ipu_dc_disable_channel(u32 dc_chan)
{
	u32 reg;
	int irq = 0, ret, timeout = 50;

	if (dc_chan == 1) {
		irq = IPU_IRQ_DC_FC_1;
	} else if (dc_chan == 5) {
		irq = IPU_IRQ_DP_SF_END;
	} else {
		return;
	}

	ret = ipu_wait_for_interrupt(irq, 50);
	if (ret)
		return;

	/* Wait for DC triple buffer to empty */
	if (dc_channels[dc_chan].di == 0)
		while ((__raw_readl(DC_STAT) & 0x00000002)
			!= 0x00000002) {
			msleep(2);
			timeout -= 2;
			if (timeout <= 0)
				break;
		}
	else if (dc_channels[dc_chan].di == 1)
		while ((__raw_readl(DC_STAT) & 0x00000020)
			!= 0x00000020) {
			msleep(2);
			timeout -= 2;
			if (timeout <= 0)
				break;
	}

	reg = __raw_readl(DC_WR_CH_CONF(dc_chan));
	reg &= ~DC_WR_CH_CONF_PROG_TYPE_MASK;
	__raw_writel(reg, DC_WR_CH_CONF(dc_chan));
}
EXPORT_SYMBOL(ipu_dc_disable_channel);

static void ipu_dc_map_config(int map, int byte_num, int offset, int mask)
{
	int ptr = map * 3 + byte_num;
	u32 reg;

	reg = __raw_readl(DC_MAP_CONF_VAL(ptr));
	reg &= ~(0xffff << (16 * (ptr & 0x1)));
	reg |= ((offset << 8) | mask) << (16 * (ptr & 0x1));
	__raw_writel(reg, DC_MAP_CONF_VAL(ptr));

	reg = __raw_readl(DC_MAP_CONF_PTR(map));
	reg &= ~(0x1f << ((16 * (map & 0x1)) + (5 * byte_num)));
	reg |= ptr << ((16 * (map & 0x1)) + (5 * byte_num));
	__raw_writel(reg, DC_MAP_CONF_PTR(map));
}

static void ipu_dc_map_clear(int map)
{
	u32 reg = __raw_readl(DC_MAP_CONF_PTR(map));
	__raw_writel(reg & ~(0xffff << (16 * (map & 0x1))),
		     DC_MAP_CONF_PTR(map));
}

int ipu_dc_init(struct platform_device *pdev, unsigned long base, unsigned long template_base)
{
	static int channel_offsets[] = { 0, 0x1c, 0x38, 0x54, 0x58, 0x5c, 0x78, 0, 0x94, 0xb4};
	int i;

	ipu_dc_reg = ioremap(base, PAGE_SIZE);
	if (!ipu_dc_reg)
		return -ENOMEM;

	ipu_dev = &pdev->dev;

	ipu_dc_tmpl_reg = ioremap(template_base, PAGE_SIZE);
	if (!ipu_dc_tmpl_reg) {
		iounmap(ipu_dc_reg);
		return -ENOMEM;
	}

	for (i = 0; i < 10; i++)
		dc_channels[i].channel_offset = channel_offsets[i];

	/* IPU_PIX_FMT_RGB24 */
	ipu_dc_map_clear(0);
	ipu_dc_map_config(0, 0, 7, 0xff);
	ipu_dc_map_config(0, 1, 15, 0xff);
	ipu_dc_map_config(0, 2, 23, 0xff);

	/* IPU_PIX_FMT_RGB666 */
	ipu_dc_map_clear(1);
	ipu_dc_map_config(1, 0, 5, 0xfc);
	ipu_dc_map_config(1, 1, 11, 0xfc);
	ipu_dc_map_config(1, 2, 17, 0xfc);

	/* IPU_PIX_FMT_YUV444 */
	ipu_dc_map_clear(2);
	ipu_dc_map_config(2, 0, 15, 0xff);
	ipu_dc_map_config(2, 1, 23, 0xff);
	ipu_dc_map_config(2, 2, 7, 0xff);

	/* IPU_PIX_FMT_RGB565 */
	ipu_dc_map_clear(3);
	ipu_dc_map_config(3, 0, 4, 0xf8);
	ipu_dc_map_config(3, 1, 10, 0xfc);
	ipu_dc_map_config(3, 2, 15, 0xf8);

	/* IPU_PIX_FMT_LVDS666 */
	ipu_dc_map_clear(4);
	ipu_dc_map_config(4, 0, 5, 0xfc);
	ipu_dc_map_config(4, 1, 13, 0xfc);
	ipu_dc_map_config(4, 2, 21, 0xfc);

	return 0;
}

void ipu_dc_exit(struct platform_device *pdev)
{
	iounmap(ipu_dc_reg);
	iounmap(ipu_dc_tmpl_reg);
}
