/*
 * Copyright (c) 2010 Sascha Hauer <s.hauer@pengutronix.de>
 * Copyright (C) 2005-2009 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/io.h>
#include <linux/mfd/imx-ipu-v3.h>
#include <linux/export.h>

#include "ipu-prv.h"

#define DMFC_RD_CHAN		0x0000
#define DMFC_WR_CHAN		0x0004
#define DMFC_WR_CHAN_DEF	0x0008
#define DMFC_DP_CHAN		0x000c
#define DMFC_DP_CHAN_DEF	0x0010
#define DMFC_GENERAL1		0x0014
#define DMFC_GENERAL2		0x0018
#define DMFC_IC_CTRL		0x001c
#define DMFC_STAT		0x0020

#define DMFC_WR_CHAN_1_28		0
#define DMFC_WR_CHAN_2_41		8
#define DMFC_WR_CHAN_1C_42		16
#define DMFC_WR_CHAN_2C_43		24

#define DMFC_DP_CHAN_5B_23		0
#define DMFC_DP_CHAN_5F_27		8
#define DMFC_DP_CHAN_6B_24		16
#define DMFC_DP_CHAN_6F_29		24

#define DMFC_FIFO_SIZE_64		(3 << 3)
#define DMFC_FIFO_SIZE_128		(2 << 3)
#define DMFC_FIFO_SIZE_256		(1 << 3)
#define DMFC_FIFO_SIZE_512		(0 << 3)

#define DMFC_SEGMENT(x)			((x & 0x7) << 0)
#define DMFC_BURSTSIZE_32		(0 << 6)
#define DMFC_BURSTSIZE_16		(1 << 6)
#define DMFC_BURSTSIZE_8		(2 << 6)
#define DMFC_BURSTSIZE_4		(3 << 6)

static struct device *ipu_dev;

struct dmfc_channel {
	int		ipu_channel;
	unsigned long	channel_reg;
	unsigned long	shift;
	unsigned	eot_shift;
	unsigned	slots;
	unsigned	max_fifo_lines;
	unsigned	slotmask;
	unsigned	segment;
};

static struct dmfc_channel dmfcs[] = {
	{
		.ipu_channel	= 23,
		.channel_reg	= DMFC_DP_CHAN,
		.shift		= DMFC_DP_CHAN_5B_23,
		.eot_shift	= 20,
		.max_fifo_lines	= 3,
	}, {
		.ipu_channel	= 24,
		.channel_reg	= DMFC_DP_CHAN,
		.shift		= DMFC_DP_CHAN_6B_24,
		.eot_shift	= 22,
		.max_fifo_lines	= 1,
	}, {
		.ipu_channel	= 27,
		.channel_reg	= DMFC_DP_CHAN,
		.shift		= DMFC_DP_CHAN_5F_27,
		.eot_shift	= 21,
		.max_fifo_lines	= 2,
	}, {
		.ipu_channel	= 28,
		.channel_reg	= DMFC_WR_CHAN,
		.shift		= DMFC_WR_CHAN_1_28,
		.eot_shift	= 16,
		.max_fifo_lines	= 2,
	}, {
		.ipu_channel	= 29,
		.channel_reg	= DMFC_DP_CHAN,
		.shift		= DMFC_DP_CHAN_6F_29,
		.eot_shift	= 23,
		.max_fifo_lines	= 1,
	},
};

#define DMFC_NUM_CHANNELS	ARRAY_SIZE(dmfcs)

static int dmfc_use_count;
static void __iomem *dmfc_regs;
static unsigned long dmfc_bandwidth_per_slot;
static DEFINE_MUTEX(dmfc_mutex);

int ipu_dmfc_enable_channel(struct dmfc_channel *dmfc)
{
	mutex_lock(&dmfc_mutex);
	ipu_get();

	if (!dmfc_use_count)
		ipu_module_enable(IPU_CONF_DMFC_EN);

	dmfc_use_count++;

	mutex_unlock(&dmfc_mutex);

	return 0;
}
EXPORT_SYMBOL(ipu_dmfc_enable_channel);

void ipu_dmfc_disable_channel(struct dmfc_channel *dmfc)
{
	mutex_lock(&dmfc_mutex);

	dmfc_use_count--;

	if (!dmfc_use_count)
		ipu_module_disable(IPU_CONF_DMFC_EN);

	if (dmfc_use_count < 0)
		dmfc_use_count = 0;

	ipu_put();
	mutex_unlock(&dmfc_mutex);
}
EXPORT_SYMBOL(ipu_dmfc_disable_channel);

static int ipu_dmfc_setup_channel(struct dmfc_channel *dmfc, int slots, int segment)
{
	u32 val, field;

	dev_dbg(ipu_dev, "dmfc: using %d slots starting from segment %d for IPU channel %d\n",
			slots, segment, dmfc->ipu_channel);

	if (!dmfc)
		return -EINVAL;

	switch (slots) {
	case 1:
		field = DMFC_FIFO_SIZE_64;
		break;
	case 2:
		field = DMFC_FIFO_SIZE_128;
		break;
	case 4:
		field = DMFC_FIFO_SIZE_256;
		break;
	case 8:
		field = DMFC_FIFO_SIZE_512;
		break;
	default:
		return -EINVAL;
	}

	field |= DMFC_SEGMENT(segment) | DMFC_BURSTSIZE_8;

	val = readl(dmfc_regs + dmfc->channel_reg);

	val &= ~(0xff << dmfc->shift);
	val |= field << dmfc->shift;

	writel(val, dmfc_regs + dmfc->channel_reg);

	dmfc->slots = slots;
	dmfc->segment = segment;
	dmfc->slotmask = ((1 << slots) - 1) << segment;

	return 0;
}

static int dmfc_bandwidth_to_slots(unsigned long bandwidth)
{
	int slots = 1;

	while (slots * dmfc_bandwidth_per_slot < bandwidth)
		slots *= 2;

	return slots;
}

static int dmfc_find_slots(int slots)
{
	unsigned slotmask_need, slotmask_used = 0;
	int i, segment = 0;

	slotmask_need = (1 << slots) - 1;

	for (i = 0; i < DMFC_NUM_CHANNELS; i++)
		slotmask_used |= dmfcs[i].slotmask;

	while (slotmask_need <= 0xff) {
		if (!(slotmask_used & slotmask_need))
			return segment;

		slotmask_need <<= 1;
		segment++;
	}

	return -EBUSY;
}

void ipu_dmfc_free_bandwidth(struct dmfc_channel *dmfc)
{
	int i;

	dev_dbg(ipu_dev, "dmfc: freeing %d slots starting from segment %d\n",
			dmfc->slots, dmfc->segment);

	mutex_lock(&dmfc_mutex);

	if (!dmfc->slots)
		goto out;

	dmfc->slotmask = 0;
	dmfc->slots = 0;
	dmfc->segment = 0;

	for (i = 0; i < ARRAY_SIZE(dmfcs); i++)
		dmfcs[i].slotmask = 0;

	for (i = 0; i < ARRAY_SIZE(dmfcs); i++) {
		if (dmfcs[i].slots > 0) {
			dmfcs[i].segment = dmfc_find_slots(dmfcs[i].slots);
			dmfcs[i].slotmask = ((1 << dmfcs[i].slots) - 1) << dmfcs[i].segment;
		}
	}

	for (i = 0; i < ARRAY_SIZE(dmfcs); i++) {
		if (dmfcs[i].slots > 0)
			ipu_dmfc_setup_channel(&dmfcs[i], dmfcs[i].slots, dmfcs[i].segment);
	}
out:
	mutex_unlock(&dmfc_mutex);
}
EXPORT_SYMBOL(ipu_dmfc_free_bandwidth);

int ipu_dmfc_alloc_bandwidth(struct dmfc_channel *dmfc,
		unsigned long bandwidth_pixel_per_second)
{
	int slots = dmfc_bandwidth_to_slots(bandwidth_pixel_per_second);
	int segment = 0, ret = 0;

	dev_dbg(ipu_dev, "dmfc: trying to allocate %ldMpixel/s for IPU channel %d\n",
			bandwidth_pixel_per_second / 1000000, dmfc->ipu_channel);

	ipu_dmfc_free_bandwidth(dmfc);

	ipu_get();
	mutex_lock(&dmfc_mutex);

	if (slots > 8) {
		ret = -EBUSY;
		goto out;
	}

	segment = dmfc_find_slots(slots);
	if (segment < 0) {
		ret = -EBUSY;
		goto out;
	}

	ipu_dmfc_setup_channel(dmfc, slots, segment);

out:
	ipu_put();
	mutex_unlock(&dmfc_mutex);

	return ret;
}
EXPORT_SYMBOL(ipu_dmfc_alloc_bandwidth);

int ipu_dmfc_init_channel(struct dmfc_channel *dmfc, int width)
{
	u32 dmfc_gen1;

	ipu_get();

	dmfc_gen1 = readl(dmfc_regs + DMFC_GENERAL1);

	if ((dmfc->slots * 64 * 4) / width > dmfc->max_fifo_lines)
		dmfc_gen1 |= 1 << dmfc->eot_shift;
	else
		dmfc_gen1 &= ~(1 << dmfc->eot_shift);

	writel(dmfc_gen1, dmfc_regs + DMFC_GENERAL1);

	ipu_put();

	return 0;
}
EXPORT_SYMBOL(ipu_dmfc_init_channel);

struct dmfc_channel *ipu_dmfc_get(int ipu_channel)
{
	int i;

	for (i = 0; i < DMFC_NUM_CHANNELS; i++)
		if (dmfcs[i].ipu_channel == ipu_channel)
			return &dmfcs[i];
	return NULL;
}
EXPORT_SYMBOL(ipu_dmfc_get);

void ipu_dmfc_put(struct dmfc_channel *dmfc)
{
	ipu_dmfc_free_bandwidth(dmfc);
}
EXPORT_SYMBOL(ipu_dmfc_put);

int ipu_dmfc_init(struct platform_device *pdev, unsigned long base,
		struct clk *ipu_clk)
{
	dmfc_regs = ioremap(base, PAGE_SIZE);

	if (!dmfc_regs)
		return -ENOMEM;

	ipu_dev = &pdev->dev;

	writel(0x0, dmfc_regs + DMFC_WR_CHAN);
	writel(0x0, dmfc_regs + DMFC_DP_CHAN);

	/*
	 * We have a total bandwidth of clkrate * 4pixel divided
	 * into 8 slots.
	 */
	dmfc_bandwidth_per_slot = clk_get_rate(ipu_clk) / 4;

	dev_dbg(ipu_dev, "dmfc: 8 slots with %ldMpixel/s bandwidth each\n",
			dmfc_bandwidth_per_slot / 1000000);

	writel(0x202020f6, dmfc_regs + DMFC_WR_CHAN_DEF);
	writel(0x2020f6f6, dmfc_regs + DMFC_DP_CHAN_DEF);
	writel(0x00000003, dmfc_regs + DMFC_GENERAL1);

	return 0;
}

void ipu_dmfc_exit(struct platform_device *pdev)
{
	iounmap(dmfc_regs);
}
