/*
 * Copyright (c) 2010 Sascha Hauer <s.hauer@pengutronix.de>
 * Copyright (C) 2005-2009 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/io.h>
#include <linux/mfd/imx-ipu-v3.h>
#include <linux/err.h>

#include "ipu-prv.h"

#define DP_SYNC 0
#define DP_ASYNC0 0x60
#define DP_ASYNC1 0xBC

#define DP_COM_CONF(flow)		(ipu_dp_base + flow)
#define DP_GRAPH_WIND_CTRL(flow)	(ipu_dp_base + 0x0004 + flow)
#define DP_FG_POS(flow)			(ipu_dp_base + 0x0008 + flow)
#define DP_CSC_A_0(flow)		(ipu_dp_base + 0x0044 + flow)
#define DP_CSC_A_1(flow)		(ipu_dp_base + 0x0048 + flow)
#define DP_CSC_A_2(flow)		(ipu_dp_base + 0x004C + flow)
#define DP_CSC_A_3(flow)		(ipu_dp_base + 0x0050 + flow)
#define DP_CSC_0(flow)			(ipu_dp_base + 0x0054 + flow)
#define DP_CSC_1(flow)			(ipu_dp_base + 0x0058 + flow)

#define DP_COM_CONF_FG_EN		(1 << 0)
#define DP_COM_CONF_GWSEL		(1 << 1)
#define DP_COM_CONF_GWAM		(1 << 2)
#define DP_COM_CONF_GWCKE		(1 << 3)
#define DP_COM_CONF_CSC_DEF_MASK	(3 << 8)
#define DP_COM_CONF_CSC_DEF_OFFSET	8
#define DP_COM_CONF_CSC_DEF_FG		(3 << 8)
#define DP_COM_CONF_CSC_DEF_BG		(2 << 8)
#define DP_COM_CONF_CSC_DEF_BOTH	(1 << 8)

struct ipu_dp {
	u32 flow;
	bool in_use;
};

static struct ipu_dp ipu_dp[3];
static struct device *ipu_dev;

static u32 ipu_flows[] = {DP_SYNC, DP_ASYNC0, DP_ASYNC1};

enum csc_type_t {
	RGB2YUV = 0,
	YUV2RGB,
	RGB2RGB,
	YUV2YUV,
	CSC_NONE,
	CSC_NUM
};

static void __iomem *ipu_dp_base;
static int dp_use_count;
static DEFINE_MUTEX(dp_mutex);

/*     Y = R *  .299 + G *  .587 + B *  .114;
       U = R * -.169 + G * -.332 + B *  .500 + 128.;
       V = R *  .500 + G * -.419 + B * -.0813 + 128.;*/
static const int rgb2ycbcr_coeff[5][3] = {
	{ 153, 301, 58, },
	{ -87, -170, 0x0100, },
	{ 0x100, -215, -42, },
	{ 0x0000, 0x0200, 0x0200, },	/* B0, B1, B2 */
	{ 0x2, 0x2, 0x2, },	/* S0, S1, S2 */
};

/*     R = (1.164 * (Y - 16)) + (1.596 * (Cr - 128));
       G = (1.164 * (Y - 16)) - (0.392 * (Cb - 128)) - (0.813 * (Cr - 128));
       B = (1.164 * (Y - 16)) + (2.017 * (Cb - 128); */
static const int ycbcr2rgb_coeff[5][3] = {
	{ 0x095, 0x000, 0x0CC, },
	{ 0x095, 0x3CE, 0x398, },
	{ 0x095, 0x0FF, 0x000, },
	{ 0x3E42, 0x010A, 0x3DD6, },	/*B0,B1,B2 */
	{ 0x1, 0x1, 0x1, },	/*S0,S1,S2 */
};

/* Please keep S0, S1 and S2 as 0x2 by using this conversion */
static int _rgb_to_yuv(int n, int red, int green, int blue)
{
	int c;

	c = red * rgb2ycbcr_coeff[n][0];
	c += green * rgb2ycbcr_coeff[n][1];
	c += blue * rgb2ycbcr_coeff[n][2];
	c /= 16;
	c += rgb2ycbcr_coeff[3][n] * 4;
	c += 8;
	c /= 16;
	if (c < 0)
		c = 0;
	if (c > 255)
		c = 255;
	return c;
}

struct dp_csc_param_t {
	int mode;
	void *coeff;
};

/*
 * Row is for BG:	RGB2YUV YUV2RGB RGB2RGB YUV2YUV CSC_NONE
 * Column is for FG:	RGB2YUV YUV2RGB RGB2RGB YUV2YUV CSC_NONE
 */
static struct dp_csc_param_t dp_csc_array[CSC_NUM][CSC_NUM] = {
	{
		{ DP_COM_CONF_CSC_DEF_BOTH, &rgb2ycbcr_coeff, },
		{ 0, 0, },
		{ 0, 0, },
		{ DP_COM_CONF_CSC_DEF_BG, &rgb2ycbcr_coeff, },
		{ DP_COM_CONF_CSC_DEF_BG, &rgb2ycbcr_coeff, },
	}, {
		{ 0, 0, },
		{ DP_COM_CONF_CSC_DEF_BOTH, &ycbcr2rgb_coeff, },
		{ DP_COM_CONF_CSC_DEF_BG, &ycbcr2rgb_coeff, },
		{ 0, 0, },
		{ DP_COM_CONF_CSC_DEF_BG, &ycbcr2rgb_coeff, },
	}, {
		{ 0, 0, },
		{ DP_COM_CONF_CSC_DEF_FG, &ycbcr2rgb_coeff, },
		{ 0, 0, },
		{ 0, 0, },
		{ 0, 0, },
	}, {
		{ DP_COM_CONF_CSC_DEF_FG, &rgb2ycbcr_coeff, },
		{ 0, 0, },
		{ 0, 0, },
		{ 0, 0, },
		{ 0, 0, },
	}, {
		{ DP_COM_CONF_CSC_DEF_FG, &rgb2ycbcr_coeff, },
		{ DP_COM_CONF_CSC_DEF_FG, &ycbcr2rgb_coeff, },
		{ 0, 0, },
		{ 0, 0, },
		{ 0, 0, },
	}
};

static enum csc_type_t fg_csc_type = CSC_NONE, bg_csc_type = CSC_NONE;

static int color_key_4rgb = 1;

int ipu_dp_set_color_key(struct ipu_dp *dp, bool enable, u32 color_key)
{
	u32 reg;
	int y, u, v;
	int red, green, blue;

	mutex_lock(&dp_mutex);

	color_key_4rgb = 1;
	/* Transform color key from rgb to yuv if CSC is enabled */
	if (((fg_csc_type == RGB2YUV) && (bg_csc_type == YUV2YUV)) ||
			((fg_csc_type == YUV2YUV) && (bg_csc_type == RGB2YUV)) ||
			((fg_csc_type == YUV2YUV) && (bg_csc_type == YUV2YUV)) ||
			((fg_csc_type == YUV2RGB) && (bg_csc_type == YUV2RGB))) {

		dev_dbg(ipu_dev, "color key 0x%x need change to yuv fmt\n", color_key);

		red = (color_key >> 16) & 0xFF;
		green = (color_key >> 8) & 0xFF;
		blue = color_key & 0xFF;

		y = _rgb_to_yuv(0, red, green, blue);
		u = _rgb_to_yuv(1, red, green, blue);
		v = _rgb_to_yuv(2, red, green, blue);
		color_key = (y << 16) | (u << 8) | v;

		color_key_4rgb = 0;

		dev_dbg(ipu_dev, "color key change to yuv fmt 0x%x\n", color_key);
	}

	if (enable) {
		reg = __raw_readl(DP_GRAPH_WIND_CTRL(dp->flow)) & 0xFF000000L;
		__raw_writel(reg | color_key, DP_GRAPH_WIND_CTRL(dp->flow));

		reg = __raw_readl(DP_COM_CONF(dp->flow));
		__raw_writel(reg | DP_COM_CONF_GWCKE, DP_COM_CONF(dp->flow));
	} else {
		reg = __raw_readl(DP_COM_CONF(dp->flow));
		__raw_writel(reg & ~DP_COM_CONF_GWCKE, DP_COM_CONF(dp->flow));
	}

	reg = ipu_cm_read(IPU_SRM_PRI2) | 0x8;
	ipu_cm_write(reg, IPU_SRM_PRI2);

	mutex_unlock(&dp_mutex);

	return 0;
}
EXPORT_SYMBOL(ipu_dp_set_color_key);

int ipu_dp_set_global_alpha(struct ipu_dp *dp, bool enable, u8 alpha, bool bg_chan)
{
	u32 reg;

	mutex_lock(&dp_mutex);

	reg = __raw_readl(DP_COM_CONF(dp->flow));
	if (bg_chan)
		reg &= ~DP_COM_CONF_GWSEL;
	else
		reg |= DP_COM_CONF_GWSEL;
	__raw_writel(reg, DP_COM_CONF(dp->flow));

	if (enable) {
		reg = __raw_readl(DP_GRAPH_WIND_CTRL(dp->flow)) & 0x00FFFFFFL;
		__raw_writel(reg | ((u32) alpha << 24),
			     DP_GRAPH_WIND_CTRL(dp->flow));

		reg = __raw_readl(DP_COM_CONF(dp->flow));
		__raw_writel(reg | DP_COM_CONF_GWAM, DP_COM_CONF(dp->flow));
	} else {
		reg = __raw_readl(DP_COM_CONF(dp->flow));
		__raw_writel(reg & ~DP_COM_CONF_GWAM, DP_COM_CONF(dp->flow));
	}

	reg = ipu_cm_read(IPU_SRM_PRI2) | 0x8;
	ipu_cm_write(reg, IPU_SRM_PRI2);

	mutex_unlock(&dp_mutex);

	return 0;
}
EXPORT_SYMBOL(ipu_dp_set_global_alpha);

int ipu_dp_set_window_pos(struct ipu_dp *dp, u16 x_pos, u16 y_pos)
{
	u32 reg;

	mutex_lock(&dp_mutex);

	__raw_writel((x_pos << 16) | y_pos, DP_FG_POS(dp->flow));

	reg = ipu_cm_read(IPU_SRM_PRI2);
	reg |= 0x8;
	ipu_cm_write(reg, IPU_SRM_PRI2);

	mutex_unlock(&dp_mutex);

	return 0;
}
EXPORT_SYMBOL(ipu_dp_set_window_pos);

#define mask_a(a) ((u32)(a) & 0x3FF)
#define mask_b(b) ((u32)(b) & 0x3FFF)

void __ipu_dp_csc_setup(int dp, struct dp_csc_param_t dp_csc_param,
			bool srm_mode_update)
{
	u32 reg;
	const int (*coeff)[5][3];

	if (dp_csc_param.mode >= 0) {
		reg = __raw_readl(DP_COM_CONF(dp));
		reg &= ~DP_COM_CONF_CSC_DEF_MASK;
		reg |= dp_csc_param.mode;
		__raw_writel(reg, DP_COM_CONF(dp));
	}

	coeff = dp_csc_param.coeff;

	if (coeff) {
		__raw_writel(mask_a((*coeff)[0][0]) |
				(mask_a((*coeff)[0][1]) << 16), DP_CSC_A_0(dp));
		__raw_writel(mask_a((*coeff)[0][2]) |
				(mask_a((*coeff)[1][0]) << 16), DP_CSC_A_1(dp));
		__raw_writel(mask_a((*coeff)[1][1]) |
				(mask_a((*coeff)[1][2]) << 16), DP_CSC_A_2(dp));
		__raw_writel(mask_a((*coeff)[2][0]) |
				(mask_a((*coeff)[2][1]) << 16), DP_CSC_A_3(dp));
		__raw_writel(mask_a((*coeff)[2][2]) |
				(mask_b((*coeff)[3][0]) << 16) |
				((*coeff)[4][0] << 30), DP_CSC_0(dp));
		__raw_writel(mask_b((*coeff)[3][1]) | ((*coeff)[4][1] << 14) |
				(mask_b((*coeff)[3][2]) << 16) |
				((*coeff)[4][2] << 30), DP_CSC_1(dp));
	}

	if (srm_mode_update) {
		reg = ipu_cm_read(IPU_SRM_PRI2) | 0x8;
		ipu_cm_write(reg, IPU_SRM_PRI2);
	}
}

int ipu_dp_setup_channel(struct ipu_dp *dp, u32 in_pixel_fmt,
		 u32 out_pixel_fmt, int bg)
{
	int in_fmt, out_fmt;
	enum csc_type_t *csc_type;
	u32 reg;

	ipu_get();

	if (bg)
		csc_type = &bg_csc_type;
	else
		csc_type = &fg_csc_type;

	in_fmt = format_to_colorspace(in_pixel_fmt);
	out_fmt = format_to_colorspace(out_pixel_fmt);

	if (in_fmt == RGB) {
		if (out_fmt == RGB)
			*csc_type = RGB2RGB;
		else
			*csc_type = RGB2YUV;
	} else {
		if (out_fmt == RGB)
			*csc_type = YUV2RGB;
		else
			*csc_type = YUV2YUV;
	}

	/* Transform color key from rgb to yuv if CSC is enabled */
	reg = __raw_readl(DP_COM_CONF(dp->flow));
	if (color_key_4rgb && (reg & DP_COM_CONF_GWCKE) &&
			(((fg_csc_type == RGB2YUV) && (bg_csc_type == YUV2YUV)) ||
			 ((fg_csc_type == YUV2YUV) && (bg_csc_type == RGB2YUV)) ||
			 ((fg_csc_type == YUV2YUV) && (bg_csc_type == YUV2YUV)) ||
			 ((fg_csc_type == YUV2RGB) && (bg_csc_type == YUV2RGB)))) {
		int red, green, blue;
		int y, u, v;
		u32 color_key = __raw_readl(DP_GRAPH_WIND_CTRL(dp->flow)) & 0xFFFFFFL;

		dev_dbg(ipu_dev, "_ipu_dp_init color key 0x%x need change to yuv fmt!\n", color_key);

		red = (color_key >> 16) & 0xFF;
		green = (color_key >> 8) & 0xFF;
		blue = color_key & 0xFF;

		y = _rgb_to_yuv(0, red, green, blue);
		u = _rgb_to_yuv(1, red, green, blue);
		v = _rgb_to_yuv(2, red, green, blue);
		color_key = (y << 16) | (u << 8) | v;

		reg = __raw_readl(DP_GRAPH_WIND_CTRL(dp->flow)) & 0xFF000000L;
		__raw_writel(reg | color_key, DP_GRAPH_WIND_CTRL(dp->flow));
		color_key_4rgb = 0;

		dev_dbg(ipu_dev, "_ipu_dp_init color key change to yuv fmt 0x%x!\n", color_key);
	}

	__ipu_dp_csc_setup(dp->flow, dp_csc_array[bg_csc_type][fg_csc_type], true);

	ipu_put();

	return 0;
}
EXPORT_SYMBOL(ipu_dp_setup_channel);

int ipu_dp_enable_channel(struct ipu_dp *dp)
{
	mutex_lock(&dp_mutex);
	ipu_get();

	if (!dp_use_count)
		ipu_module_enable(IPU_CONF_DP_EN);

	dp_use_count++;

	mutex_unlock(&dp_mutex);

	return 0;
}
EXPORT_SYMBOL(ipu_dp_enable_channel);

void ipu_dp_disable_channel(struct ipu_dp *dp)
{
	mutex_lock(&dp_mutex);

	dp_use_count--;

	if (!dp_use_count)
		ipu_module_disable(IPU_CONF_DP_EN);

	if (dp_use_count < 0)
		dp_use_count = 0;

	ipu_put();
	mutex_unlock(&dp_mutex);
}
EXPORT_SYMBOL(ipu_dp_disable_channel);

void ipu_dp_enable_fg(struct ipu_dp *dp)
{
	u32 reg;

	/* Enable FG channel */
	reg = __raw_readl(DP_COM_CONF(DP_SYNC));
	__raw_writel(reg | DP_COM_CONF_FG_EN, DP_COM_CONF(DP_SYNC));

	reg = ipu_cm_read(IPU_SRM_PRI2);
	reg |= 0x8;
	ipu_cm_write(reg, IPU_SRM_PRI2);
}
EXPORT_SYMBOL(ipu_dp_enable_fg);

void ipu_dp_disable_fg(struct ipu_dp *dp)
{
	u32 reg, csc;

	ipu_get();

	reg = __raw_readl(DP_COM_CONF(DP_SYNC));
	csc = reg & DP_COM_CONF_CSC_DEF_MASK;
	if (csc == DP_COM_CONF_CSC_DEF_FG)
		reg &= ~DP_COM_CONF_CSC_DEF_MASK;

	reg &= ~DP_COM_CONF_FG_EN;
	__raw_writel(reg, DP_COM_CONF(DP_SYNC));

	reg = ipu_cm_read(IPU_SRM_PRI2) | 0x8;
	ipu_cm_write(reg, IPU_SRM_PRI2);

	ipu_put();
}
EXPORT_SYMBOL(ipu_dp_disable_fg);

struct ipu_dp *ipu_dp_get(unsigned int flow)
{
	struct ipu_dp *dp;

	if (flow > 2)
		return ERR_PTR(-EINVAL);

	dp = &ipu_dp[flow];

	if (dp->in_use)
		return ERR_PTR(-EBUSY);

	dp->in_use = true;
	dp->flow = ipu_flows[flow];

	return dp;
}
EXPORT_SYMBOL(ipu_dp_get);

void ipu_dp_put(struct ipu_dp *dp)
{
	dp->in_use = false;
}
EXPORT_SYMBOL(ipu_dp_put);

int ipu_dp_init(struct platform_device *pdev, unsigned long base)
{
	ipu_dp_base = ioremap(base, PAGE_SIZE);
	if (!ipu_dp_base)
		return -ENOMEM;

	ipu_dev = &pdev->dev;

	return 0;
}

void ipu_dp_exit(struct platform_device *pdev)
{
	 iounmap(ipu_dp_base);
}
